#include <string.h>
#include <algorithm> // std::sort
#include <numeric>   // std::accumulate

#include "cdo_remap.h"


const auto is_not_equal = [](auto a, auto b) noexcept { return  (a < b || b < a); };
const auto is_equal = [](auto a, auto b) noexcept { return !(a < b || b < a); };

extern "C"
{
#include "clipping.h"
#include "area.h"
#include "geometry.h"
}

#define AREA_TOL_FACTOR (1e-6)

struct CellSearch
{
  enum yac_edge_type *edgeType = nullptr;
  size_t numCellCorners = 0;
  size_t maxCells = 0;
  Varray<double> overlapAreas;
  Varray<struct grid_cell> gridCells;
  Varray<struct grid_cell> overlapCells;
};

static void
cellsearch_realloc(size_t numCells, CellSearch &search)
{
  if (numCells > search.maxCells)
    {
      search.overlapAreas.resize(numCells);
      search.overlapCells.resize(numCells);
      search.gridCells.resize(numCells);

      for (size_t i = search.maxCells; i < numCells; ++i)
        {
          search.overlapCells[i].array_size = 0;
          search.overlapCells[i].num_corners = 0;
          search.overlapCells[i].edge_type = nullptr;
          search.overlapCells[i].coordinates_xyz = nullptr;

          search.gridCells[i].array_size = search.numCellCorners;
          search.gridCells[i].num_corners = search.numCellCorners;
          search.gridCells[i].edge_type = search.edgeType;
          search.gridCells[i].coordinates_xyz = new double[search.numCellCorners][3];
        }

      search.maxCells = numCells;
    }
}

static void
cellsearch_free(CellSearch &search)
{
  for (size_t i = 0; i < search.maxCells; i++)
    {
      if (search.overlapCells[i].array_size > 0)
        {
          if (search.overlapCells[i].coordinates_xyz) free(search.overlapCells[i].coordinates_xyz);
          if (search.overlapCells[i].edge_type) free(search.overlapCells[i].edge_type);
        }

      delete[] search.gridCells[i].coordinates_xyz;
    }
}

static enum yac_cell_type
get_cell_type(struct grid_cell target_cell)
{
  /*
  YAC_ASSERT(
    target_cell.num_corners > 0, "ERROR(get_cell_type): CELL without edge")
  */
  enum yac_cell_type cell_type = MIXED_CELL;

  // if the cell is a typical lon-lat cell
  if ((target_cell.num_corners == 4) &&
      ((target_cell.edge_type[0] == LAT_CIRCLE_EDGE &&
        target_cell.edge_type[1] == LON_CIRCLE_EDGE &&
        target_cell.edge_type[2] == LAT_CIRCLE_EDGE &&
        target_cell.edge_type[3] == LON_CIRCLE_EDGE) ||
       (target_cell.edge_type[0] == LON_CIRCLE_EDGE &&
        target_cell.edge_type[1] == LAT_CIRCLE_EDGE &&
        target_cell.edge_type[2] == LON_CIRCLE_EDGE &&
        target_cell.edge_type[3] == LAT_CIRCLE_EDGE))) {

    cell_type = LON_LAT_CELL;

  } else {

    size_t count_lat_edges = 0, count_great_circle_edges = 0;

    // count the number of edges for each type
    // (lon edges are counted as gc ones)
    for (size_t i = 0; i < target_cell.num_corners; ++i)
      if (target_cell.edge_type[i] == LON_CIRCLE_EDGE ||
          target_cell.edge_type[i] == GREAT_CIRCLE_EDGE)
        count_great_circle_edges++;
      else
        count_lat_edges++;

    // if the cell is a lon lat cell with one lat edge having a length of zero
    // due to being at a pole
    if ((count_lat_edges == 1) && (count_great_circle_edges == 2)) {

      size_t i;
      for (i = 0; i < 3; ++i) if (target_cell.edge_type[i] == LAT_CIRCLE_EDGE) break;
      size_t pol_index = (3 + i - 1) % 3;
      if (fabs(fabs(target_cell.coordinates_xyz[pol_index][2])-1.0) <
          yac_angle_low_tol) cell_type = LON_LAT_CELL;

    // if the cell only consists of great circle edges
    } else if (count_lat_edges == 0) cell_type = GREAT_CIRCLE_CELL;

    // if the cell only consists of lat circle edges
    else if (count_great_circle_edges == 0) cell_type = LAT_CELL;
  }
  /*
  YAC_ASSERT(
    cell_type != MIXED_CELL,
    "invalid cell type (cell contains edges consisting "
    "of great circles and circles of latitude)")
  */
  return cell_type;
}

static inline double
gridcell_area(const grid_cell &cell)
{
  return yac_huiliers_area(cell);
}

static void
cdo_compute_overlap_areas(size_t numSearchCells, CellSearch &search, const GridCell &gridCell)
{
  auto &overlapCells = search.overlapCells;

  // Do the clipping and get the cell for the overlapping area
  yac_cell_clipping(numSearchCells, search.gridCells.data(), gridCell.yacGridCell, overlapCells.data());

  // Get the partial areas for the overlapping regions
  for (size_t i = 0; i < numSearchCells; i++) search.overlapAreas[i] = gridcell_area(overlapCells[i]);

#ifdef VERBOSE
  for (size_t i = 0; i < numSearchCells; i++) printf("overlap area : %lf\n", search.overlapAreas[i]);
#endif
}

static void
cdo_compute_concave_overlap_info(size_t numSearchCells, CellSearch &search, const GridCell &gridCell, double tgtLon, double tgtLat,
                                 double *overlapAreas, double (*overlapBarycenters)[3])
{
  auto &overlapCells = search.overlapCells;
  auto &sourceCells = search.gridCells;
  const auto targetCell = gridCell.yacGridCell;

  if (overlapBarycenters != NULL)
    for (size_t i = 0; i < numSearchCells; ++i)
      for (int j = 0; j < 3; ++j)
        overlapBarycenters[i][j] = 0.0;

  const auto target_cell_type = (targetCell.num_corners > 3) ? get_cell_type(targetCell) : MIXED_CELL;

  if (targetCell.num_corners < 4 || target_cell_type == LON_LAT_CELL)
    {
      yac_cell_clipping(numSearchCells, sourceCells.data(), targetCell, overlapCells.data());
      for (size_t i = 0; i < numSearchCells; ++i)
        {
          if (overlapCells[i].num_corners > 1)
            {
              if (overlapBarycenters == NULL)
                overlapAreas[i] = yac_huiliers_area(overlapCells[i]);
              else
                {
                  overlapAreas[i] = yac_huiliers_area_info(overlapCells[i], overlapBarycenters[i]);
                  normalise_vector(overlapBarycenters[i]);
                }
            }
          else
            {
              overlapAreas[i] = 0.0;
            }
        }
      return;
    }

  double coordinates_xyz[3][3] = { { -1.0, -1.0, -1.0 }, { -1.0, -1.0, -1.0 }, { -1.0, -1.0, -1.0 } };
  enum yac_edge_type edgeTypes[3] = { GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE, GREAT_CIRCLE_EDGE };
  
  struct grid_cell targetPartialCell;
  targetPartialCell.array_size = 3;
  targetPartialCell.num_corners = 3;
  targetPartialCell.coordinates_xyz = coordinates_xyz;
  targetPartialCell.edge_type = edgeTypes;

  // Do the clipping and get the cell for the overlapping area

  for (size_t i = 0; i < numSearchCells; i++) overlapAreas[i] = 0.0;

  // common node point to all partial target cells
  auto partial_cell_xyz = targetPartialCell.coordinates_xyz;
  auto cell_xyz = targetCell.coordinates_xyz;

  LL_to_XYZ(tgtLon, tgtLat, partial_cell_xyz[0]);

  for (size_t cornerNum = 0; cornerNum < targetCell.num_corners; ++cornerNum)
    {
      const auto cornerA = cornerNum;
      const auto cornerB = (cornerNum + 1) % targetCell.num_corners;

      // skip clipping and area calculation for degenerated triangles
      //
      // If this is not sufficient, instead we can try something like:
      //
      //     point_list target_list
      //     init_point_list(&target_list);
      //     generate_point_list(&target_list, targetCell);
      //     grid_cell temp_target_cell;
      //     generate_overlap_cell(target_list, temp_target_cell);
      //     free_point_list(&target_list);
      //
      // and use temp_target_cell for triangulation.
      //
      // Compared to the if statement below the alternative seems to be quite costly.

      if (points_are_identically(cell_xyz[cornerA], cell_xyz[cornerB])) continue;

      partial_cell_xyz[1][0] = cell_xyz[cornerA][0];
      partial_cell_xyz[1][1] = cell_xyz[cornerA][1];
      partial_cell_xyz[1][2] = cell_xyz[cornerA][2];
      partial_cell_xyz[2][0] = cell_xyz[cornerB][0];
      partial_cell_xyz[2][1] = cell_xyz[cornerB][1];
      partial_cell_xyz[2][2] = cell_xyz[cornerB][2];

      yac_cell_clipping(numSearchCells, sourceCells.data(), targetPartialCell, overlapCells.data());

      // Get the partial areas for the overlapping regions as sum over the partial target cells.
      for (size_t i = 0; i < numSearchCells; i++)
        {
          if (overlapCells[i].num_corners == 0) continue;

          if (overlapBarycenters == NULL)
            overlapAreas[i] += gridcell_area(overlapCells[i]);
          else
            overlapAreas[i] += yac_huiliers_area_info(overlapCells[i], overlapBarycenters[i]);
        }
    }

  if (overlapBarycenters != NULL)
    for (size_t i = 0; i < numSearchCells; i++)
      if (overlapAreas[i] > 0.0)
        normalise_vector(overlapBarycenters[i]);

#ifdef VERBOSE
  for (size_t i = 0; i < numSearchCells; i++) printf("overlap area %zu: %lf\n", i, overlapAreas[i]);
#endif
}

static void
cdo_compute_concave_overlap_areas(size_t numSearchCells, CellSearch &search, const GridCell &gridCell, double tgtLon, double tgtLat)
{
  auto overlapAreas = search.overlapAreas.data();
  cdo_compute_concave_overlap_info(numSearchCells, search, gridCell, tgtLon, tgtLat, overlapAreas, NULL);
}

static void
set_cell_coordinates_yac(const size_t cellIndex, const size_t numCorners, const GridInfo &gridInfo,
                         const struct grid_cell &yacGridCell, double *x, double *y)
{
  const auto storeXY = (x && y);
  auto xyz = yacGridCell.coordinates_xyz;

  const auto cellCornersLon = &gridInfo.cellCornersLon[cellIndex * numCorners];
  const auto cellCornersLat = &gridInfo.cellCornersLat[cellIndex * numCorners];
  for (size_t i = 0; i < numCorners; ++i) LL_to_XYZ(cellCornersLon[i], cellCornersLat[i], xyz[i]);

  if (storeXY)
    {
      for (size_t k = 0; k < numCorners; ++k)
        {
          x[k] = cellCornersLon[k];
          y[k] = cellCornersLat[k];
        }
    }
}

static void
set_cell_coordinates(const size_t cellIndex, const size_t numCorners, const GridInfo &gridInfo, const GridCell &gridCell)
{
  set_cell_coordinates_yac(cellIndex, numCorners, gridInfo, gridCell.yacGridCell, gridCell.coordinates_x, gridCell.coordinates_y);
}

static void
set_coordinates_yac(const size_t numCells, const Varray<size_t> &cellIndices,
                    const size_t numCorners, const GridInfo &gridInfo, const Varray<grid_cell> &yacGridCell)
{
  for (size_t i = 0; i < numCells; ++i)
    set_cell_coordinates_yac(cellIndices[i], numCorners, gridInfo, yacGridCell[i], nullptr, nullptr);
}

static void
set_coordinates_yac(const double (*xyzCoords)[3], const size_t numCells, const Varray<size_t> &cellIndices,
                    const size_t numCorners, const Varray<grid_cell> &yacGridCell)
{
  for (size_t i = 0; i < numCells; ++i)
    {
      const auto offset = cellIndices[i] * numCorners;
      auto xyz = yacGridCell[i].coordinates_xyz;
      for (size_t k = 0; k < numCorners; ++k)
        for (size_t l = 0; l < 3; ++l)
          xyz[k][l] = xyzCoords[offset + k][l];
    }
}
 
static void
gridcell_init_yac(GridCell &gridCell, size_t numCorners, enum yac_edge_type *edgeType)
{
  gridCell.yacGridCell.array_size = numCorners;
  gridCell.yacGridCell.num_corners = numCorners;
  gridCell.yacGridCell.edge_type = edgeType;
  gridCell.yacGridCell.coordinates_xyz = new double[numCorners][3];
  gridCell.coordinates_x = new double[numCorners];
  gridCell.coordinates_y = new double[numCorners];
}

static void
gridcell_free_yac(const GridCell &gridCell)
{
  delete[] gridCell.yacGridCell.coordinates_xyz;
  delete[] gridCell.coordinates_x;
  delete[] gridCell.coordinates_y;
}

static size_t
remove_invalid_areas(size_t numSearchCells, Varray<double> &areas, Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numSearchCells; ++i)
    {
      if (areas[i] > 0.0)
        {
          areas[n] = areas[i];
          searchIndices[n] = searchIndices[i];
          n++;
        }
    }

  return n;
}

static size_t
remove_invalid_weights(size_t gridSize, size_t numWeights, Varray<double> &weights, Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numWeights; ++i)
    {
      const auto cellIndex = (weights[i] > 0.0) ? searchIndices[i] : gridSize;
      if (cellIndex != gridSize)
        {
          weights[n] = weights[i];
          searchIndices[n] = cellIndex;
          n++;
        }
    }

  return n;
}

static size_t
remove_unmask_weights(const Varray<short> &gridMask, size_t numWeights, Varray<double> &weights,
                      Varray<size_t> &searchIndices)
{
  size_t n = 0;
  for (size_t i = 0; i < numWeights; ++i)
    {
      const auto cellIndex = searchIndices[i];
      /*
        Store the appropriate addresses and weights.
        Also add contributions to cell areas.
        The source grid mask is the master mask.
      */
      if (gridMask[cellIndex])
        {
          weights[n] = weights[i];
          searchIndices[n] = cellIndex;
          n++;
        }
    }

  return n;
}

static void
normalize_weights(NormOpt normOpt, double cellArea, double cellFrac, size_t numWeights, Varray<double> &weights)
{
  if (normOpt == NormOpt::DESTAREA)
    {
      const auto normFactor = is_not_equal(cellArea, 0.0) ? 1.0 / cellArea : 0.0;
      for (size_t i = 0; i < numWeights; ++i) weights[i] *= normFactor;
    }
  else if (normOpt == NormOpt::FRACAREA)
    {
      const auto normFactor = is_not_equal(cellFrac, 0.0) ? 1.0 / cellFrac : 0.0;
      for (size_t i = 0; i < numWeights; ++i) weights[i] *= normFactor;
    }
  else if (normOpt == NormOpt::NONE)
    {
    }
}

static void
correct_weights(double cellArea, size_t numWeights, Varray<double> &weights)
{
  for (size_t i = 0; i < numWeights; ++i) weights[i] /= cellArea;
  yac_correct_weights(numWeights, weights.data());
  for (size_t i = 0; i < numWeights; ++i) weights[i] *= cellArea;
}

static void
sort_weights(size_t numWeights, Varray<size_t> &searchIndices, Varray<double> &weights)
{
  size_t n;
  for (n = 1; n < numWeights; ++n)
    if (searchIndices[n] < searchIndices[n - 1]) break;
  if (n == numWeights) return;

  if (numWeights > 1)
    {
      struct IndexWeightX
      {
        size_t index;
        double weight;
      };

      Varray<IndexWeightX> indexWeights(numWeights);

      for (n = 0; n < numWeights; ++n)
        {
          indexWeights[n].index = searchIndices[n];
          indexWeights[n].weight = weights[n];
        }

      const auto compareIndex = [](const auto &a, const auto &b) noexcept { return a.index < b.index; };
      std::sort(indexWeights.begin(), indexWeights.end(), compareIndex);

      for (n = 0; n < numWeights; ++n)
        {
          searchIndices[n] = indexWeights[n].index;
          weights[n] = indexWeights[n].weight;
        }
    }
}

static int
get_lonlat_circle_index(size_t numCells, size_t numCorners, const Varray<double> &clon, const Varray<double> &clat)
{
  int lonlatCircleIndex = -1;

  if (numCorners == 4)
    {
      size_t iadd = (numCells < 100) ? 1 : numCells / 30 - 1;
      size_t num_i = 0, num_eq0 = 0, num_eq1 = 0;

      for (size_t i = 0; i < numCells; i += iadd)
        {
          const auto i4 = i * 4;
          num_i++;
          // clang-format off
          if (is_equal(clon[i4 + 1], clon[i4 + 2]) && is_equal(clon[i4 + 3], clon[i4 + 0]) &&
              is_equal(clat[i4 + 0], clat[i4 + 1]) && is_equal(clat[i4 + 2], clat[i4 + 3]))
            {
              num_eq1++;
            }
          else if (is_equal(clon[i4 + 0], clon[i4 + 1]) && is_equal(clon[i4 + 2], clon[i4 + 3]) &&
                   is_equal(clat[i4 + 1], clat[i4 + 2]) && is_equal(clat[i4 + 3], clat[i4 + 0]))
            {
              num_eq0++;
            }
          // clang-format on
        }

      if (num_i == num_eq1) lonlatCircleIndex = 1;
      if (num_i == num_eq0) lonlatCircleIndex = 0;
    }

  // printf("lonlatCircleIndex %d\n", lonlatCircleIndex);

  return lonlatCircleIndex;
}

static int
get_lonlat_circle_index(const GridInfo &gridInfo)
{
  return get_lonlat_circle_index(gridInfo.numCells, gridInfo.numCorners, gridInfo.cellCornersLon, gridInfo.cellCornersLat);
}

static double
conserv_remap(const Varray<double> &srcArray, size_t numWeights, const Varray<double> &weights, const Varray<size_t> &srcIndices)
{
  double tgtPoint = 0.0;
  for (size_t i = 0; i < numWeights; ++i) tgtPoint += srcArray[srcIndices[i]] * weights[i];

  return tgtPoint;
}

void
remap_conserv(NormOpt normOpt, const RemapSearch &remapSearch, const Varray<double> &srcArray, Varray<double> &tgtArray, double missval)
{
  auto &srcGrid = remapSearch.srcGrid;
  auto &tgtGrid = remapSearch.tgtGrid;

  auto srcNumCells = srcGrid.numCells;
  auto tgtNumCells = tgtGrid.numCells;

  Varray<short> srcGridMask(srcNumCells);
  for (size_t i = 0; i < srcNumCells; ++i) srcGridMask[i] = !is_equal(srcArray[i], missval);

  auto srcNumCorners = srcGrid.numCorners;
  auto tgtNumCorners = tgtGrid.numCorners;

  enum yac_edge_type lonlatCircleType[] = { LON_CIRCLE_EDGE, LAT_CIRCLE_EDGE, LON_CIRCLE_EDGE, LAT_CIRCLE_EDGE, LON_CIRCLE_EDGE };
  Varray<enum yac_edge_type> greatCircleType(std::max(srcNumCorners, tgtNumCorners), GREAT_CIRCLE_EDGE);

  auto srcEdgeType = greatCircleType.data();
  auto tgtEdgeType = greatCircleType.data();

  enum yac_cell_type tgtCellType = MIXED_CELL;

  if (srcNumCorners == 4)
    {
      const auto lonlatCircleIndex = get_lonlat_circle_index(srcGrid);
      if (lonlatCircleIndex >= 0) srcEdgeType = &lonlatCircleType[lonlatCircleIndex];
    }

  if (tgtNumCorners == 4)
    {
      const auto lonlatCircleIndex = get_lonlat_circle_index(tgtGrid);
      if (lonlatCircleIndex >= 0)
        {
          tgtCellType = LON_LAT_CELL;
          tgtEdgeType = &lonlatCircleType[lonlatCircleIndex];
        }
    }

  GridCell tgtGridCell;
  gridcell_init_yac(tgtGridCell, tgtNumCorners, tgtEdgeType);

  CellSearch cellSearch;
  cellSearch.numCellCorners = srcNumCorners;
  cellSearch.edgeType = srcEdgeType;

  auto numCellCorners = srcNumCorners;  // num of corners of search cells

  Varray<size_t> searchIndices(srcNumCells);

  // Loop over target grid
  for (size_t tgtCellIndex = 0; tgtCellIndex < tgtNumCells; ++tgtCellIndex)
    {
      tgtArray[tgtCellIndex] = missval;

      //if (!tgtGrid.mask[tgtCellIndex]) continue;

      set_cell_coordinates(tgtCellIndex, tgtNumCorners, tgtGrid, tgtGridCell);

      // Get search cells
      auto numSearchCells = do_gridcell_search(remapSearch.gcs, false, tgtGridCell, searchIndices);

      if (numSearchCells == 0) continue;

      // Create search arrays

      cellsearch_realloc(numSearchCells, cellSearch);

      if (remapSearch.gcs.xyzCoords)
        set_coordinates_yac(remapSearch.gcs.xyzCoords, numSearchCells, searchIndices, numCellCorners, cellSearch.gridCells);
      else
        set_coordinates_yac(numSearchCells, searchIndices, numCellCorners, srcGrid, cellSearch.gridCells);

      if (tgtNumCorners < 4 || tgtCellType == LON_LAT_CELL)
        {
          cdo_compute_overlap_areas(numSearchCells, cellSearch, tgtGridCell);
        }
      else
        {
          cdo_compute_concave_overlap_areas(numSearchCells, cellSearch, tgtGridCell,
                                            tgtGrid.cellCentersLon[tgtCellIndex], tgtGrid.cellCentersLat[tgtCellIndex]);
        }

      auto &partialWeights = cellSearch.overlapAreas;

      auto numWeights = remove_invalid_areas(numSearchCells, partialWeights, searchIndices);

      const auto tgtCellArea = gridcell_area(tgtGridCell.yacGridCell);

      if (normOpt == NormOpt::FRACAREA) correct_weights(tgtCellArea, numWeights, partialWeights);

      numWeights = remove_invalid_weights(srcNumCells, numWeights, partialWeights, searchIndices);
      numWeights = remove_unmask_weights(srcGridMask, numWeights, partialWeights, searchIndices);

      const auto tgtCellFrac = std::accumulate(partialWeights.begin(), partialWeights.begin() + numWeights, 0.0);

      if (numWeights)
        {
          sort_weights(numWeights, searchIndices, partialWeights);
          // Normalize weights using target cell area if requested
          normalize_weights(normOpt, tgtCellArea, tgtCellFrac, numWeights, partialWeights);
          tgtArray[tgtCellIndex] = conserv_remap(srcArray, numWeights, partialWeights, searchIndices);
        }
    }

  // Finished with all cells: deallocate search arrays

  cellsearch_free(cellSearch);
  gridcell_free_yac(tgtGridCell);
}  // remap_conserv




struct weight_vector_data {
  double weight;
  size_t local_id;
  size_t global_id;
};

struct weight_vector_data_3d {
  double weight[3];
  size_t local_id;
  size_t global_id;
};

struct weight_vector_3d {
  struct weight_vector_data_3d *data;
  size_t n;
};

struct supermesh_cell {
  struct {
    size_t local_id;
    size_t global_id;
  } src, tgt;
  double norm_area;
  double area;
  double barycenter[3];
  struct weight_vector_3d * src_cell_gradient;
};


static int
compare_supermesh_cell_src_local_ids(const void * a, const void * b)
{
  struct supermesh_cell * a_ = (struct supermesh_cell *)a;
  struct supermesh_cell * b_ = (struct supermesh_cell *)b;

  int ret = (a_->src.local_id > b_->src.local_id) -
            (a_->src.local_id < b_->src.local_id);
  if (ret) return ret;
  return (a_->tgt.global_id > b_->tgt.global_id) -
         (a_->tgt.global_id < b_->tgt.global_id);
}

static int
compare_supermesh_cell_tgt_local_ids(const void * a, const void * b)
{
  struct supermesh_cell * a_ = (struct supermesh_cell *)a;
  struct supermesh_cell * b_ = (struct supermesh_cell *)b;

  int ret = (a_->tgt.local_id > b_->tgt.local_id) -
            (a_->tgt.local_id < b_->tgt.local_id);
  if (ret) return ret;
  return (a_->src.global_id > b_->src.global_id) -
         (a_->src.global_id < b_->src.global_id);
}

static inline void
orthogonalise_weight_vector(double M[3][3], struct weight_vector_3d * G_i, struct weight_vector_data_3d * buffer)
{

  // This routine computes: (I_3 - C_i * C_i^-1) * G_i
  //
  // O(g_i) = g_i - C_i * (C_i^-1 * g_i)
  // where: C_i is the centeroid of a source cell
  //        g_i is the gradient of the source field in C_i
  //        O(g_i) is the projection of g_i into the plane perpendicular to C_i
  // g_i = G_i * f
  // where: G_i is the weight matrix to compute g_i
  //        f is the source field vector
  // => O(g_i) = (I_3 - C_i * C_i^-1) * G_i * f
  // where: I_3 is the identity matrix of size 3 x 3
  //
  // M = I_3 - C_i * C_i^-1

  struct weight_vector_data_3d * G_i_data = G_i->data;

  size_t N = G_i->n;
  for (size_t i = 0; i < N; ++i) {
    buffer[i].local_id = G_i_data[i].local_id;
    buffer[i].global_id = G_i_data[i].global_id;
    for (size_t j = 0; j < 3; ++j) buffer[i].weight[j] = 0.0;
  }

  for (size_t n = 0; n < N; ++n)
    for (size_t i = 0; i < 3; ++i)
      for (size_t j = 0; j < 3; ++j)
        buffer[n].weight[i] += G_i_data[n].weight[j] * M[i][j];

  memcpy(G_i->data, buffer, N * sizeof(*buffer));
}

static int
compare_weight_vector_data_weight(void const * a, void const * b)
{
  struct weight_vector_data const * weight_a = (struct weight_vector_data const *)a;
  struct weight_vector_data const * weight_b = (struct weight_vector_data const *)b;

  int ret = weight_a->global_id - weight_b->global_id;
  if (ret) return ret;
  double abs_weight_a = fabs(weight_a->weight);
  double abs_weight_b = fabs(weight_b->weight);
  ret = (abs_weight_a > abs_weight_b) - (abs_weight_a < abs_weight_b);
  if (ret) return ret;
  return (weight_a->weight > weight_b->weight) -
         (weight_a->weight < weight_b->weight);
}

static int
compare_weight_vector_data(void const * a, void const * b)
{
  struct weight_vector_data const * weight_a = (struct weight_vector_data const *)a;
  struct weight_vector_data const * weight_b = (struct weight_vector_data const *)b;

  return weight_a->global_id - weight_b->global_id;
}

static void
compact_weight_vector_data(struct weight_vector_data * weights, size_t * n)
{
  size_t n_ = *n;

  if (n_ <= 1) return;

  // sort weights by global_id then by weight
  qsort(weights, n_, sizeof(*weights), compare_weight_vector_data_weight);

  size_t new_n = 1;
  struct weight_vector_data * prev_weight_data = weights;
  struct weight_vector_data * curr_weight_data = weights + 1;
  for (size_t i = 1; i < n_; ++i, ++curr_weight_data) {

    // if both weights refer to the same source point (by global_id)
    if (!compare_weight_vector_data(prev_weight_data, curr_weight_data)) {
      prev_weight_data->weight += curr_weight_data->weight;
    } else {
      ++new_n;
      ++prev_weight_data;
      *prev_weight_data = *curr_weight_data;
    }
  }

  *n = new_n;
}
/*
static void get_cell_buffers_(
  struct interp_grid * interp_grid,
  struct grid_cell * tgt_grid_cell, struct grid_cell * src_grid_cell)
{
  struct const_basic_grid_data * src_basic_grid_data =
    yac_interp_grid_get_basic_grid_data_src(interp_grid);
  struct const_basic_grid_data * tgt_basic_grid_data =
    yac_interp_grid_get_basic_grid_data_tgt(interp_grid);

  int max_num_vertices_per_cell =
    MAX(get_max_num_vertices_per_cell(src_basic_grid_data),
        get_max_num_vertices_per_cell(tgt_basic_grid_data));

  enum yac_edge_type * edge_type_buffer =
    xmalloc(2 * (size_t)max_num_vertices_per_cell *
            sizeof(*edge_type_buffer));
  coordinate_pointer coordinates_xyz_buffer =
    xmalloc(2 * (size_t)max_num_vertices_per_cell *
            sizeof(*coordinates_xyz_buffer));

  tgt_grid_cell->coordinates_xyz = coordinates_xyz_buffer;
  tgt_grid_cell->edge_type = edge_type_buffer;
  tgt_grid_cell->array_size = max_num_vertices_per_cell;

  src_grid_cell->coordinates_xyz = coordinates_xyz_buffer + max_num_vertices_per_cell;
  src_grid_cell->edge_type = edge_type_buffer + max_num_vertices_per_cell;
  src_grid_cell->array_size = max_num_vertices_per_cell;
}
*/
static coordinate_pointer
compute_src_cell_centroids(const double (*xyzCoords)[3], const CellSearch &search,
  struct interp_grid * interp_grid,
  size_t *src_cells, int *skip_src_cell, size_t num_src_cells,
  struct supermesh_cell *super_cells, size_t num_super_cells)
{
  coordinate_pointer src_cell_centroids = (coordinate_pointer)malloc(num_src_cells * sizeof(*src_cell_centroids));

  // sort supermesh cells first by source local id and second by target global id
  qsort(super_cells, num_super_cells, sizeof(*super_cells), compare_supermesh_cell_src_local_ids);

  struct grid_cell src_grid_cell;
  const auto numCorners = search.numCellCorners;
  src_grid_cell.coordinates_xyz = new double[numCorners][3];
  src_grid_cell.edge_type = search.edgeType;
  src_grid_cell.num_corners = numCorners;
  src_grid_cell.array_size = numCorners;

  // struct const_basic_grid_data * src_basic_grid_data = yac_interp_grid_get_basic_grid_data_src(interp_grid);

  // compute centroids of source cells
  // C_i = N(S_(U_k) (A_k*C_k))
  // where: N(C) = C*(C*C)^-0.5 // normalisation
  //        U_k all supermesh cells overlaping with the respective source cell
  //        A_k area of supermesh cell
  //        C_k barycenter of supermesh cell (normalisation of the sum of all
  //                                          vertices of the cell)
  for (size_t i = 0, offset = 0; i < num_src_cells; ++i) {

    if (skip_src_cell[i]) continue;

    size_t curr_src_cell = src_cells[i];
    //yac_const_basic_grid_data_get_grid_cell(src_basic_grid_data, curr_src_cell, &src_grid_cell);
    const auto xoffset = curr_src_cell * numCorners;
    auto xyz = src_grid_cell.coordinates_xyz;
    for (size_t k = 0; k < numCorners; ++k)
      for (size_t l = 0; l < 3; ++l)
        xyz[k][l] = xyzCoords[(xoffset + k)][l];
    double src_cell_area = yac_huiliers_area(src_grid_cell);

    struct supermesh_cell * curr_super_cells = super_cells + offset;
    size_t curr_num_super_cells = offset;
    while ((offset < num_super_cells) &&
           (super_cells[offset].src.local_id == curr_src_cell)) ++offset;
    curr_num_super_cells = offset - curr_num_super_cells;

    double src_cell_area_diff = src_cell_area;
    double src_cell_centroid[3] = {0.0, 0.0, 0.0};

    for (size_t j = 0; j < curr_num_super_cells; ++j) {

      double super_cell_area = curr_super_cells[j].area;
      double * super_cell_barycenter = curr_super_cells[j].barycenter;
      src_cell_centroid[0] += super_cell_area * super_cell_barycenter[0];
      src_cell_centroid[1] += super_cell_area * super_cell_barycenter[1];
      src_cell_centroid[2] += super_cell_area * super_cell_barycenter[2];
      src_cell_area_diff -= super_cell_area;
    }

    if (!(skip_src_cell[i] =
          (fabs(src_cell_area_diff) > src_cell_area * AREA_TOL_FACTOR))) {
    
      normalise_vector(src_cell_centroid);
      src_cell_centroids[i][0] = src_cell_centroid[0];
      src_cell_centroids[i][1] = src_cell_centroid[1];
      src_cell_centroids[i][2] = src_cell_centroid[2];
    }
  }
  delete[] src_grid_cell.coordinates_xyz;
  //free(src_grid_cell.edge_type);

  return src_cell_centroids;
}

static void
compute_cell_barycenter(struct const_basic_grid_data * grid_data, size_t cell_idx, double barycenter[3])
{
  size_t num_vertices = grid_data->num_vertices_per_cell[cell_idx];
  size_t const *vertices = grid_data->cell_to_vertex + grid_data->cell_to_vertex_offsets[cell_idx];

  barycenter[0] = 0.0;
  barycenter[1] = 0.0;
  barycenter[2] = 0.0;

  for (size_t i = 0; i < num_vertices; ++i) {
    double const *curr_vertex_coordinate = grid_data->vertex_coordinates[vertices[i]];
    barycenter[0] += curr_vertex_coordinate[0];
    barycenter[1] += curr_vertex_coordinate[1];
    barycenter[2] += curr_vertex_coordinate[2];
  }
  normalise_vector(barycenter);
}

static struct weight_vector_3d *
compute_src_cell_gradients(
  struct interp_grid * interp_grid, size_t * src_cells,
  coordinate_pointer src_cell_centroids, int * skip_src_cell,
  size_t num_src_cells, size_t * src_cell_neighbours)
{
  struct weight_vector_3d *src_cell_gradients = (struct weight_vector_3d *)malloc(num_src_cells * sizeof(*src_cell_gradients));

  //struct const_basic_grid_data * src_basic_grid_data = yac_interp_grid_get_basic_grid_data_src(interp_grid);

  size_t total_num_gradient_weights = 0;
  size_t max_num_neigh_per_src = 0;
  for (size_t i = 0, offset = 0; i < num_src_cells; ++i) {
    src_cell_gradients[i].data = NULL;
    src_cell_gradients[i].n = 0;
    size_t curr_num_neigh = src_basic_grid_data->num_vertices_per_cell[src_cells[i]];
    size_t * curr_neighs = src_cell_neighbours + offset;
    offset += curr_num_neigh;
    for (size_t j = 0; (!skip_src_cell[i]) && (j < curr_num_neigh); ++j)
      skip_src_cell[i] |= curr_neighs[j] == SIZE_MAX;
    if (skip_src_cell[i]) continue;
    total_num_gradient_weights += curr_num_neigh + 1;
    if (max_num_neigh_per_src < curr_num_neigh)
      max_num_neigh_per_src = curr_num_neigh;
  }
  struct weight_vector_data_3d *weight_vector_data_buffer =
    (total_num_gradient_weights > 0) ?
    ((struct weight_vector_data_3d *)malloc(total_num_gradient_weights * sizeof(*weight_vector_data_buffer))) : NULL;
  for (size_t i = 0; i < total_num_gradient_weights; ++i)
    for (size_t j = 0; j < 3; ++j)
      weight_vector_data_buffer[i].weight[j] = 0.0;

  struct weight_vector_data_3d *orth_buffer = (struct weight_vector_data_3d *)malloc((max_num_neigh_per_src + 1) * sizeof(*orth_buffer));

  // compute gradient in the centroid for each source cell
  // g_i = O_i(g'_i)
  // where: O_i(g) = g - C_i*(C_i*g) // makes g orthogonal to C_i
  // g'_i = (A_(C_k))^-1 * S_ijk(((f_j + f_k) * 0.5 - f_i) *
  //                         (C_j x C_k) * |C_j x C_k|^-1 *
  //                         asin(|C_j x C_k|))
  // where: g'_i is the estimated centroid gradient
  //        A_(C_k) is the area of the cell generated by connecting the
  //                barycenters of all neighbour cells
  //        S_ijk is the sum of all edges of the previously described cell in
  //              counterclockwise order
  //        f_i, f_j, and f_k are the mean values of the field over the area of
  //                          the respective source cell area
  // asin(|C_j x C_k|) / |C_j x C_k| ~ 1.0
  // => g'_i = S_ijk(((f_j + f_k) * 0.5 - f_i) * (C_j x C_k)) / A_(C_k)
  //
  // g_i = G_i * f
  // G_i = S_ijk(((I_j + I_k) * 0.5 - I_i) * (C_j x C_k)) / A_(C_k)
  // where: G_i is the gradient weight matrix
  //        I_x is a vector of the same size as f, which is all zero except at
  //            position x, where it is one
  //        f is the source field vector
  for (size_t i = 0, offset = 0, weight_vector_data_buffer_offset = 0; i < num_src_cells; ++i)
    {
    size_t curr_num_neigh = src_basic_grid_data->num_vertices_per_cell[src_cells[i]];
    size_t * curr_neighs = src_cell_neighbours + offset;
    offset += curr_num_neigh;
    struct weight_vector_3d * G_i = src_cell_gradients + i;
    G_i->data = weight_vector_data_buffer + weight_vector_data_buffer_offset;

    if (skip_src_cell[i]) continue;

    weight_vector_data_buffer_offset += curr_num_neigh + 1;
    G_i->n = curr_num_neigh + 1;
    G_i->data[0].local_id = src_cells[i];
    G_i->data[0].global_id = src_basic_grid_data->cell_ids[src_cells[i]];
    for (size_t j = 0; j < curr_num_neigh; ++j)
      {
        G_i->data[j+1].local_id = curr_neighs[j];
        G_i->data[j+1].global_id = src_basic_grid_data->cell_ids[curr_neighs[j]];
      }

    // area of the polygon that is formed by connecting the barycenters of
    // the neighbouring cells
    double A_C_k = 0.0;

    struct grid_cell centroid_triangle = {
      .coordinates_xyz = (double[3][3]){{0}},
      .edge_type = (enum yac_edge_type[]) {GREAT_CIRCLE_EDGE,GREAT_CIRCLE_EDGE,GREAT_CIRCLE_EDGE},
      .num_corners = 3, .array_size = 0};
    centroid_triangle.coordinates_xyz[0][0] = src_cell_centroids[i][0];
    centroid_triangle.coordinates_xyz[0][1] = src_cell_centroids[i][1];
    centroid_triangle.coordinates_xyz[0][2] = src_cell_centroids[i][2];

    // the following matrix is used by the routine
    // orthogonalise_weight_vector
    // M = I_3 - C_i * C_i^-1
    double M[3][3];
    for (size_t k = 0; k < 3; ++k)
      for (size_t l = 0; l < 3; ++l)
        M[k][l] = - centroid_triangle.coordinates_xyz[0][k] *
                    centroid_triangle.coordinates_xyz[0][l];
    for (size_t k = 0; k < 3; ++k)
      M[k][k] += 1.0;

    // We split the cell that is comprised of the barycenters of the edge
    // neigbours into triangles, which has the centroid of the current cell
    // as one corner. The sum of the areas of these triangles is A_C_K.
    for (size_t j = 0; j < curr_num_neigh; ++j) {

      size_t curr_neighbours[2];
      curr_neighbours[0] = curr_neighs[j];
      curr_neighbours[1] = curr_neighs[(j+1)%curr_num_neigh];

      double neigh_cell_barycenters[2][3];
      compute_cell_barycenter(src_basic_grid_data, curr_neighbours[0], neigh_cell_barycenters[0]);
      compute_cell_barycenter(src_basic_grid_data, curr_neighbours[1], neigh_cell_barycenters[1]);

      centroid_triangle.coordinates_xyz[1][0] = neigh_cell_barycenters[0][0];
      centroid_triangle.coordinates_xyz[1][1] = neigh_cell_barycenters[0][1];
      centroid_triangle.coordinates_xyz[1][2] = neigh_cell_barycenters[0][2];
      centroid_triangle.coordinates_xyz[2][0] = neigh_cell_barycenters[1][0];
      centroid_triangle.coordinates_xyz[2][1] = neigh_cell_barycenters[1][1];
      centroid_triangle.coordinates_xyz[2][2] = neigh_cell_barycenters[1][2];

      A_C_k += yac_huiliers_area(centroid_triangle);

      // C_j x C_k
      double C_j_x_C_k[3];
      crossproduct_ld(
        neigh_cell_barycenters[0], neigh_cell_barycenters[1], C_j_x_C_k);
      // this factor ensures that C_j_x_C_k points outwards of the centroid
      // cell
      double factor =
        -copysign(1.0, C_j_x_C_k[0] * centroid_triangle.coordinates_xyz[0][0] +
                       C_j_x_C_k[1] * centroid_triangle.coordinates_xyz[0][1] +
                       C_j_x_C_k[2] * centroid_triangle.coordinates_xyz[0][2]);

      for (size_t l = 0; l < 3; ++l) C_j_x_C_k[l] *= factor;

      // -I_i * (C_j x C_k)
      G_i->data[0].weight[0] -= C_j_x_C_k[0];
      G_i->data[0].weight[1] -= C_j_x_C_k[1];
      G_i->data[0].weight[2] -= C_j_x_C_k[2];

      for (size_t l = 0; l < 3; ++l) C_j_x_C_k[l] *= 0.5;

      // 0.5 * I_j * (C_j x C_k)
      G_i->data[j+1].weight[0] += C_j_x_C_k[0];
      G_i->data[j+1].weight[1] += C_j_x_C_k[1];
      G_i->data[j+1].weight[2] += C_j_x_C_k[2];

      // 0.5 * I_k * (C_j x C_k)
      size_t j_ = (j + 1) % curr_num_neigh + 1;
      G_i->data[j_].weight[0] += C_j_x_C_k[0];
      G_i->data[j_].weight[1] += C_j_x_C_k[1];
      G_i->data[j_].weight[2] += C_j_x_C_k[2];
    }

    // ((I_j + I_k) * 0.5 - I_i) * (C_j x C_k) / A_(C_k)
    double inv_A_C_K = 1.0 / A_C_k;
    for (size_t k = 0; k <= curr_num_neigh; ++k)
      for (size_t l = 0; l < 3; ++l)
        G_i->data[k].weight[l] *= inv_A_C_K;

    orthogonalise_weight_vector(M, G_i, orth_buffer);
  }
  free(orth_buffer);

  return src_cell_gradients;
}
 */
void
remap_conserv2(NormOpt normOpt, const RemapSearch &remapSearch, const Varray<double> &srcArray, Varray<double> &tgtArray, double missval)
{
  auto &srcGrid = remapSearch.srcGrid;
  auto &tgtGrid = remapSearch.tgtGrid;

  auto srcNumCells = srcGrid.numCells;
  auto tgtNumCells = tgtGrid.numCells;

  Varray<short> srcGridMask(srcNumCells);
  for (size_t i = 0; i < srcNumCells; ++i) srcGridMask[i] = !is_equal(srcArray[i], missval);

  auto srcNumCorners = srcGrid.numCorners;
  auto tgtNumCorners = tgtGrid.numCorners;

  enum yac_edge_type lonlatCircleType[] = { LON_CIRCLE_EDGE, LAT_CIRCLE_EDGE, LON_CIRCLE_EDGE, LAT_CIRCLE_EDGE, LON_CIRCLE_EDGE };
  Varray<enum yac_edge_type> greatCircleType(std::max(srcNumCorners, tgtNumCorners), GREAT_CIRCLE_EDGE);

  auto srcEdgeType = greatCircleType.data();
  auto tgtEdgeType = greatCircleType.data();

  enum yac_cell_type tgtCellType = MIXED_CELL;

  if (srcNumCorners == 4)
    {
      const auto lonlatCircleIndex = get_lonlat_circle_index(srcGrid);
      if (lonlatCircleIndex >= 0) srcEdgeType = &lonlatCircleType[lonlatCircleIndex];
    }

  if (tgtNumCorners == 4)
    {
      const auto lonlatCircleIndex = get_lonlat_circle_index(tgtGrid);
      if (lonlatCircleIndex >= 0)
        {
          tgtCellType = LON_LAT_CELL;
          tgtEdgeType = &lonlatCircleType[lonlatCircleIndex];
        }
    }

  GridCell tgtGridCell;
  gridcell_init_yac(tgtGridCell, tgtNumCorners, tgtEdgeType);

  CellSearch cellSearch;
  cellSearch.numCellCorners = srcNumCorners;
  cellSearch.edgeType = srcEdgeType;

  auto numCellCorners = srcNumCorners;  // num of corners of search cells

  Varray<size_t> searchIndices(srcNumCells);

  // Loop over target grid
  for (size_t tgtCellIndex = 0; tgtCellIndex < tgtNumCells; ++tgtCellIndex)
    {
      tgtArray[tgtCellIndex] = missval;

      //if (!tgtGrid.mask[tgtCellIndex]) continue;

      set_cell_coordinates(tgtCellIndex, tgtNumCorners, tgtGrid, tgtGridCell);

      // Get search cells
      auto numSearchCells = do_gridcell_search(remapSearch.gcs, false, tgtGridCell, searchIndices);

      if (numSearchCells == 0) continue;

      // Create search arrays

      cellsearch_realloc(numSearchCells, cellSearch);

      set_coordinates_yac(numSearchCells, searchIndices, numCellCorners, srcGrid, cellSearch.gridCells);

      // For all supermesh cells compute the area and normalised area.
      // Additionally, remove all empty supermesh cells.
      size_t new_num_super_cells = 0;
      size_t tgt_idx = 0;
      /*
      for (size_t i = 0, j = 0; i < total_num_overlaps;)
        {

          // get information about the current target cell
          size_t curr_tgt_cell = super_cells[i].tgt.local_id;
          yac_const_basic_grid_data_get_grid_cell(tgt_basic_grid_data, curr_tgt_cell, &tgt_grid_cell);

          double curr_tgt_cell_coverage = 0.0;

          // for all supermesh cells overlapping with the current target cell
          for (;(i < total_num_overlaps) && (super_cells[i].tgt.local_id == curr_tgt_cell); ++i)
            {

              // get the current source cell
              yac_const_basic_grid_data_get_grid_cell(src_basic_grid_data, super_cells[i].src.local_id, &src_grid_cell);

              // compute area of the current supermesh cell
              double super_cell_area;
              double barycenter[3];
              cdo_compute_concave_overlap_info(1, cellSearch, tgtGridCell,
                                               tgtGrid.cellCentersLon[tgtCellIndex], tgtGrid.cellCentersLat[tgtCellIndex],
                                               &super_cell_area, &barycenter);
            }

          // if there is an overlap between the current source and target cell
          if (super_cell_area > 0.0)
            {
              super_cells[new_num_super_cells].src = super_cells[i].src;
              super_cells[new_num_super_cells].tgt = super_cells[i].tgt;
              super_cells[new_num_super_cells].area = super_cell_area;
              memcpy(super_cells[new_num_super_cells].barycenter, barycenter, 3 * sizeof(double));
              super_cells[new_num_super_cells].src_cell_gradient = NULL;
              ++new_num_super_cells;

              curr_tgt_cell_coverage += super_cell_area;
            }
        }
      */
      auto &partialWeights = cellSearch.overlapAreas;

      auto numWeights = remove_invalid_areas(numSearchCells, partialWeights, searchIndices);

      const auto tgtCellArea = gridcell_area(tgtGridCell.yacGridCell);

      if (normOpt == NormOpt::FRACAREA) correct_weights(tgtCellArea, numWeights, partialWeights);

      numWeights = remove_invalid_weights(srcNumCells, numWeights, partialWeights, searchIndices);
      numWeights = remove_unmask_weights(srcGridMask, numWeights, partialWeights, searchIndices);

      const auto tgtCellFrac = std::accumulate(partialWeights.begin(), partialWeights.begin() + numWeights, 0.0);

      if (numWeights)
        {
          sort_weights(numWeights, searchIndices, partialWeights);
          // Normalize weights using target cell area if requested
          normalize_weights(normOpt, tgtCellArea, tgtCellFrac, numWeights, partialWeights);
          tgtArray[tgtCellIndex] = conserv_remap(srcArray, numWeights, partialWeights, searchIndices);
        }
    }

  // Finished with all cells: deallocate search arrays

  cellsearch_free(cellSearch);
  gridcell_free_yac(tgtGridCell);
}  // remap_conserv2
